"""Registry for things that require consent from a user."""

from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _

from djblets.privacy.consent.errors import (ConsentRequirementConflictError,
                                            ConsentRequirementNotFoundError)
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         ATTRIBUTE_REGISTERED,
                                         DEFAULT_ERRORS,
                                         OrderedRegistry,
                                         UNREGISTER)


_registry = None


class ConsentRequirementsRegistry(OrderedRegistry):
    """A registry for managing aspects of a product requiring consent.

    Each requirement in the registry requires consent by the user before the
    action or processing for that requirement is allowed to occur.
    """

    lookup_attrs = ('requirement_id',)
    already_registered_error_class = ConsentRequirementConflictError
    lookup_error_class = ConsentRequirementNotFoundError

    default_errors = dict(DEFAULT_ERRORS, **{
        ALREADY_REGISTERED: _(
            'Could not register consent requirement %(item)s: This '
            'requirement is already registered or its ID conflicts with '
            'another consent requirement.',
        ),
        ATTRIBUTE_REGISTERED: _(
            'Could not register consent requirement %(item)s: Another '
            'requirement (%(duplicate)s) is already registered with the '
            'same ID.',
        ),
        UNREGISTER: _(
            'Could not unregister consent requirement %(item)s: This '
            'requirement was not yet registered.',
        ),
    })

    def get_consent_requirement(self, requirement_id):
        """Return a consent requirement with the given ID.

        Args:
            requirement_id (unicode):
                The consent requirement ID to look up.

        Returns:
            djblets.privacy.consent.base.BaseConsentRequirement:
            The resulting consent requirement, if found. If a requirement
            with this ID could not be found, this will return ``None``.
        """
        try:
            return self.get('requirement_id', requirement_id)
        except ConsentRequirementNotFoundError:
            return None


def get_consent_requirements_registry():
    """Return the registry for managing consent requirements.

    Returns:
        ConsentRequirementsRegistry:
        The registry for consent requirements.
    """
    global _registry

    if _registry is None:
        _registry = ConsentRequirementsRegistry()

    return _registry



#
# errors.py -- Extension errors.
#
# Copyright (c) 2010-2011  Beanbag, Inc.
# Copyright (c) 2008-2010  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from django.utils.translation import ugettext as _


class EnablingExtensionError(Exception):
    """An extension could not be enabled."""

    def __init__(self, message, load_error=None, needs_reload=False):
        """Initialize the error.

        Args:
            message (unicode):
                The detailed error message.

            load_error (unicode, optional):
                An exception from the attempt to enable the extension, or
                other useful information to display to the user to help
                diagnose the problem.

            needs_reload (bool, optional):
                Whether fixing this error requires reloading the extension.
        """
        super(EnablingExtensionError, self).__init__(message)

        self.load_error = load_error
        self.needs_reload = needs_reload


class DisablingExtensionError(Exception):
    """An extension could not be disabled."""
    pass


class InstallExtensionError(Exception):
    """An extension could not be installed."""
    def __init__(self, message, load_error=None):
        self.message = message
        self.load_error = load_error


class InstallExtensionMediaError(InstallExtensionError):
    """An error indicating that extension media files could not be installed.
    """


class InvalidExtensionError(Exception):
    """An extension does not exist."""
    def __init__(self, extension_id):
        super(InvalidExtensionError, self).__init__()
        self.message = _("Cannot find extension with id %s") % extension_id



"""General utility functions for working with e-mail."""

from __future__ import unicode_literals

from email.utils import formataddr, parseaddr

from django.conf import settings


def build_email_address(email, full_name=None):
    """Build an e-mail address for a To/CC/BCC field from a user's information.

    Args:
        email (unicode):
            The e-mail address.

        full_name (unicode, optional):
            The optional full name associated with the e-mail address.

    Returns:
        unicode:
        A formatted e-mail address intended for a To/CC/BCC field.
    """
    return formataddr((full_name, email))


def build_email_address_for_user(user):
    """Build an e-mail address for a To/CC/BCC field from a User.

    Args:
        user (django.contrib.auth.models.User):
            The user.

    Returns:
        unicode:
        A formatted e-mail address intended for a To/CC/BCC field.
    """
    return build_email_address(email=user.email,
                               full_name=user.get_full_name())


def build_email_address_via_service(email, full_name=None, service_name=None,
                                    sender_email=None):
    """Build an e-mail address for sending on behalf of a user via a service.

    This will construct a formatted e-mail address that can be safely used
    in a :mailheader:`From` field without risking being quarantined/rejected
    by DMARC rules.

    The address will be in the form of "Full Name via Service Name
    <sender@domain.tld>".

    Args:
        email (unicode):
            The unformatted e-mail address of the user.

        full_name (unicode, optional):
            The full name of the user. If not provided, the username in the
            e-mail address will be used.

        service_name (unicode, optional):
            The name of the service sending the e-mail. If not provided,
            ``settings.EMAIL_DEFAULT_SENDER_SERVICE_NAME`` will be used.

        sender_email (unicode, optional):
            The unformatted e-mail address for the sending service. If not
            provided, the e-mail address in
            :django:setting:`DEFAULT_FROM_EMAIL` will be used.

    Returns:
        unicode:
        A formatted e-mail address safe to use in a :mailheader:`From` field.
    """
    if not service_name:
        # A service name wasn't specified. We'll try to use the one from
        # settings, and if that doesn't exist, we'll use the domain name
        # from the sender (assuming it parsed, and if it didn't, there are
        # bigger problems we're not going to deal with here).
        service_name = (
            getattr(settings, 'EMAIL_DEFAULT_SENDER_SERVICE_NAME', None) or
            email.split('@')[-1]
        )

    if not sender_email:
        sender_email = parseaddr(settings.DEFAULT_FROM_EMAIL)[1]

    # We need a name from the user. If a full name wasn't
    # available, use the first part of the e-mail address.
    if not full_name:
        full_name = email.split('@')[0]

    return build_email_address(
        email=sender_email,
        full_name='%s via %s' % (full_name, service_name))



from __future__ import unicode_literals


class NeedsIntegrationManagerMixin(object):
    """Mixin for classes that require an integration manager to be defined.

    Some classes require access to an integration manager, which must be
    provided by the consuming application. This mixin provides
    :py:meth:`get_integration_manager` class method that, if not overridden,
    will raise a :py:exc:`NotImplementedError`.
    """

    @classmethod
    def get_integration_manager(cls):
        """Return the integration manager used for this class.

        Subclasses must override this to return an
        :py:class:`~djblets.integrations.manager.IntegrationManager` instance
        that will be used for any interaction management for this class.

        Returns:
            djblets.integrations.manager.IntegrationManager:
            The integration manager for this class.

        Raises:
            NotImplementedError:
                Raised if a subclass doesn't override this method.
        """
        raise NotImplementedError('%r must implement get_integration_manager()'
                                  % cls)



from __future__ import unicode_literals

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.http import HttpResponse, Http404
from django.utils import six
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from django.views.generic.base import TemplateView
from django.views.generic.edit import FormView

from djblets.integrations.mixins import NeedsIntegrationManagerMixin
from djblets.util.templatetags.djblets_images import build_srcset


class IntegrationListContextViewMixin(NeedsIntegrationManagerMixin):
    """A mixin for views that display lists of integrations.

    This allows for custom views that display details on the available
    integrations and all configurations.
    """

    def get_integration_js_view_data(self):
        """Return data for a JavaScript view for the page.

        This will include the list of available integrations IDs, a mapping of
        integration IDs to details, and a list of all configurations.

        Returns:
            dict:
            The data for the JavaScript view.
        """
        integration_manager = self.get_integration_manager()
        integrations_map = {}
        integration_ids = []
        configs = []

        for integration in integration_manager.get_integrations():
            integration_ids.append(integration.integration_id)
            integrations_map[integration.integration_id] = {
                'addURL': self.get_add_config_url(integration),
                'description': integration.description,
                'iconSrc': integration.icon_static_urls['1x'],
                'iconSrcSet': build_srcset(integration.icon_static_urls),
                'id': integration.integration_id,
                'name': integration.name,
            }

        queryset = (
            self.get_configs_queryset()
            .only('enabled', 'integration_id', 'pk', 'name')
        )

        for config in queryset:
            integration_id = config.integration_id

            if integration_id in integrations_map:
                configs.append({
                    'editURL': self.get_edit_config_url(config),
                    'enabled': config.enabled,
                    'id': config.pk,
                    'integrationID': integration_id,
                    'name': config.name,
                })

        return {
            'configs': sorted(
                configs,
                key=lambda config:
                    (integrations_map[config['integrationID']]['name'],
                     config['name'])),
            'csrfToken': self.request.META['CSRF_COOKIE'],
            'integrationIDs': sorted(
                integration_ids,
                key=lambda integration_id:
                    integrations_map[integration_id]['name']),
            'integrationsMap': integrations_map,
        }

    def get_add_config_url(self, integration):
        """Return the URL for adding a new configuration.

        This can be overridden by subclasses to return a URL for another
        namespace or to add additional keyword arguments for the lookup.

        Args:
            integration (djblets.integrations.integration.Integration):
                The integration to add configurations for.

        Returns:
            unicode:
            The Add Configuration URL for the integration.
        """
        return reverse(
            'integration-add-config',
            kwargs={
                'integration_id': integration.integration_id,
            })

    def get_edit_config_url(self, config):
        """Return the URL for editing a configuration.

        This can be overridden by subclasses to return a URL for another
        namespace or to add additional keyword arguments for the lookup.

        Args:
            config (djblets.integrations.models.BaseIntegrationConfig):
                The configuration to return the URL for.

        Returns:
            unicode:
            The URL for editing the configuration.
        """
        return reverse(
            'integration-change-config',
            kwargs={
                'integration_id': config.integration.integration_id,
                'config_id': config.pk,
            })

    def get_configs_queryset(self):
        """Return a queryset for integration configs.

        Subclasses can override this to provide a more strict query to filter
        down the configurations.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for fetching integration configurations.
        """
        return self.get_integration_manager().config_model.objects.all()


class BaseIntegrationListView(IntegrationListContextViewMixin,
                              TemplateView):
    """Base class for a view that lists available integrations.

    This view handles the display of all available integrations, along with
    any existing configurations.

    This is meant to be subclassed to either fine-tune the queries for
    configurations (for instance, by limiting to a particular user or
    organization) or to add access control.
    """

    #: The name of the template used for the page.
    #:
    #: This must be provided if subclassing this yourself.
    template_name = None

    @method_decorator(login_required)
    @method_decorator(csrf_protect)
    def dispatch(self, *args, **kwargs):
        """Handle the request to the view.

        This will first check to make sure the user is logged in.

        Args:
            *args (tuple):
                Arguments to pass to the view.

            **kwargs (dict):
                Keyword arguments to pass to the view.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response.
        """
        return super(BaseIntegrationListView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        """Return context data for the template.

        By default, this returns a dictionary with a sole ``integrations``
        key, which contains data on each available integration and all
        saved configurations.

        Args:
            **kwargs (dict):
                Any arguments captured in the URL.

        Returns:
            dict:
            A dictionary of context data for the template.
        """
        context = super(BaseIntegrationListView, self).get_context_data(
            **kwargs)
        context['integrationViewData'] = self.get_integration_js_view_data()

        return context


class BaseAdminIntegrationListView(BaseIntegrationListView):
    """Base class for an admin view that lists available integrations.

    This builds upon :py:class:`BaseIntegrationListView`, adding access
    checks to ensure that only administrators can access it.
    """

    template_name = 'integrations/admin/integration_list.html'

    @method_decorator(staff_member_required)
    def dispatch(self, *args, **kwargs):
        """Handle the request to the view.

        This will first check to make sure the user is logged in and is a
        staff member.

        Args:
            *args (tuple):
                Arguments to pass to the view.

            **kwargs (dict):
                Keyword arguments to pass to the view.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response.
        """
        return super(BaseAdminIntegrationListView, self).dispatch(
            *args, **kwargs)


class BaseIntegrationConfigFormView(NeedsIntegrationManagerMixin,
                                    FormView):
    """Base class for a view that manages an integration configuration.

    This view handles the display of a form for either creating a new
    integration configuration or updating an existing one.

    This is meant to be subclassed to either fine-tune the queries for
    configurations (for instance, by limiting to a particular user or
    organization) or to add access control.
    """

    #: The name of the template used for the page.
    #:
    #: This must be provided if subclassing this yourself.
    template_name = None

    @method_decorator(login_required)
    @method_decorator(csrf_protect)
    def dispatch(self, request, *args, **kwargs):
        """Handle the request to the view.

        This will first check to make sure the user is logged in. It then
        looks up the appropriate configuration ID from the URL before passing
        it on to the view.

        Args:
            *args (tuple):
                Positional arguments to pass to the view.

            **kwargs (dict):
                Keyword arguments to pass to the view.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response.

        Raises:
            django.http.Http404:
                The integration or configuration with the given ID was not
                found.
        """
        integration_mgr = self.get_integration_manager()
        self.integration_mgr = integration_mgr

        self.integration = integration_mgr.get_integration(
            kwargs['integration_id'])

        if not self.integration:
            raise Http404

        if 'config_id' in kwargs:
            try:
                self.config = integration_mgr.get_integration_configs(
                    integration_cls=self.integration.__class__,
                    pk=kwargs['config_id'],
                    **self.get_config_query_kwargs(**kwargs))[0]
            except IndexError:
                raise Http404
        else:
            self.config = None

        return super(BaseIntegrationConfigFormView, self).dispatch(
            request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        """Handle HTTP DELETE requests.

        This will delete the integration configuration.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments passed to the view.

            **args (dict):
                Keyword arguments passed to the view.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response.
        """
        self.config.delete()

        return HttpResponse(status=204)

    def get_config_query_kwargs(self, **kwargs):
        """Return query arguments for fetching an integration configuration.

        This can be subclassed to return additional arguments used when
        fetching configurations, based on the needs of the application. For
        example, limiting it by user or organization.

        By default, this doesn't return any additional query arguments.

        Args:
            **kwargs (dict):
                Any arguments captured in the URL.

        Returns:
            dict:
            Additional query arguments as a dictionary. This will be turned
            into keyword arguments for a filter query.
        """
        return {}

    def get_context_data(self, **kwargs):
        """Return context data for the configuration page.

        Args:
            **kwargs (dict):
                Additional keyword arguments that may be passed to this
                function by the parent views.

        Returns:
            dict:
            The context to provide on the page.
        """
        data = super(BaseIntegrationConfigFormView, self).get_context_data(
            **kwargs)
        data['success_url'] = self.get_success_url()

        return data

    def get_form_kwargs(self):
        """Return keyword arguments to pass to the form.

        This will, by default, provide ``integration`` and configuration
        ``instance`` keyword arguments to the form during initialization,
        along with the ``request``.

        Subclases can override it with additional arguments if needed.

        Returns:
            dict:
            A dictionary of keyword arguments to pass to the form.
        """
        form_kwargs = \
            super(BaseIntegrationConfigFormView, self).get_form_kwargs()
        form_kwargs.update({
            'integration': self.integration,
            'request': self.request,
            'instance': self.config,
        })

        return form_kwargs

    def get_success_url(self):
        """Return the URL to redirect to when successfully saving the form.

        This defaults to returning back to the integrations page. Consumers
        that have special values to fill out in the URL will need to
        override this.

        Returns:
            unicode:
            The URL to redirect to.
        """
        return reverse('integration-list')

    def get_form_class(self):
        """Return the class used for the configuration form.

        This will return whatever class is specified for that integration.

        This function is used internally by Django's generic views framework.
        It should not be overridden.

        Returns:
            djblets.integrations.forms.IntegrationConfigForm:
            The form sublass used for integration configuration.
        """
        return self.integration.config_form_cls

    def form_valid(self, form):
        """Handle the saving of a valid configuration form

        This will save the configuration and then perform a redirect to the
        success URL, defined by :py:meth:`get_success_url`.

        Args:
            form (djblets.integrations.forms.IntegrationConfigForm):
                The form to save.

        Returns:
            django.http.HttpResponse:
            An HTTP response redirecting to the success URL.
        """
        self.object = form.save()

        return super(BaseIntegrationConfigFormView, self).form_valid(form)


class BaseAdminIntegrationConfigFormView(BaseIntegrationConfigFormView):
    """Base class for an admin view that manages an integration configuration.

    This builds upon :py:class:`BaseIntegrationConfigFormView`, adding access
    checks to ensure that only administrators can access it.
    """

    template_name = 'integrations/admin/configure_integration.html'

    @method_decorator(staff_member_required)
    def dispatch(self, *args, **kwargs):
        """Handle the request to the view.

        This will first check to make sure the user is logged in and is a
        staff member.

        Args:
            *args (tuple):
                Arguments to pass to the view.

            **kwargs (dict):
                Keyword arguments to pass to the view.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response.
        """
        return super(BaseAdminIntegrationConfigFormView, self).dispatch(
            *args, **kwargs)

    def get_context_data(self, **kwargs):
        """Return context data for the configuration page.

        Args:
            **kwargs (dict):
                Additional keyword arguments that may be passed to this
                function by the parent views.

        Returns:
            dict:
            The context to provide on the page.
        """
        data = (
            super(BaseAdminIntegrationConfigFormView, self)
            .get_context_data(**kwargs)
        )
        data.update({
            'opts': self.integration_mgr.config_model._meta,
            'original': self.config,
            'show_delete_link': self.config is not None,
            'show_save': True,
        })

        return data



"""Base test case for privacy consent tests."""

from __future__ import unicode_literals

from django.core.cache import cache

from djblets.privacy.consent import get_consent_requirements_registry
from djblets.privacy.consent.tracker import clear_consent_tracker
from djblets.testing.testcases import TestCase


class ConsentTestCase(TestCase):
    """Base test case for consent-related unit tests."""

    def setUp(self):
        super(ConsentTestCase, self).setUp()

        self.clear_consent_caches()

    def tearDown(self):
        super(ConsentTestCase, self).tearDown()

        self.clear_consent_caches()

    def clear_consent_caches(self):
        """Clear all consent-related caches."""
        cache.clear()
        get_consent_requirements_registry().reset()
        clear_consent_tracker()



"""Unit tests for djblets.webapi.resources.mixins.forms.UpdateFormMixin."""

from django import forms
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from kgb import SpyAgency

from djblets.testing.testcases import TestCase
from djblets.webapi.errors import INVALID_FORM_DATA
from djblets.webapi.resources.mixins.forms import UpdateFormMixin


class MyForm(forms.ModelForm):
    first_name = forms.CharField(initial='initial-first-name',
                                 required=False)

    def __init__(self, *args, **kwargs):
        self.custom = kwargs.pop('custom', False)

        super(MyForm, self).__init__(*args, **kwargs)

    class Meta:
        model = User
        fields = ['username', 'password', 'first_name', 'last_name']


class UpdateFormMixinTests(SpyAgency, TestCase):
    """Unit tests for djblets.webapi.resources.mixins.forms.UpdateFormMixin."""

    def setUp(self):
        super(UpdateFormMixinTests, self).setUp()

        self.mixin = UpdateFormMixin()
        self.mixin.form_class = MyForm
        self.mixin.item_result_key = 'user'

    def test_build_form_errors_with_form(self):
        """Testing UpdateFormMixin.build_form_errors with form"""
        form = MyForm(data={})

        self.assertEqual(
            self.mixin.build_form_errors(form=form),
            {
                'username': ['This field is required.'],
                'password': ['This field is required.'],
            })

    def test_build_form_errors_with_validation_error_dict(self):
        """Testing UpdateFormMixin.build_form_errors with ValidationError
        using dictionary
        """
        errors = ValidationError({
            'field1': ['This field is required.'],
            'field2': ['This is not a valid choice.'],
        })

        self.assertEqual(
            self.mixin.build_form_errors(errors=errors),
            {
                'field1': ['This field is required.'],
                'field2': ['This is not a valid choice.'],
            })

    def test_build_form_errors_with_validation_error_list(self):
        """Testing UpdateFormMixin.build_form_errors with ValidationError
        using list
        """
        errors = ValidationError([
            'Oh no.',
            ValidationError('Everything went wrong.'),
        ])

        self.assertEqual(
            self.mixin.build_form_errors(errors=errors),
            {
                '__all__': ['Oh no.', 'Everything went wrong.'],
            })

    def test_build_form_success_response_with_is_created_true(self):
        """Testing UpdateFormMixin.build_form_success_response with
        is_created=True
        """
        form = MyForm()
        user = User(username='my-user')
        response = self.mixin.build_form_success_response(form=form,
                                                          instance=user,
                                                          is_created=True)

        self.assertEqual(
            response,
            (201, {
                'user': user,
            }))

    def test_build_form_success_response_with_is_created_false(self):
        """Testing UpdateFormMixin.build_form_success_response with
        is_created=False
        """
        form = MyForm()
        user = User(username='my-user')
        response = self.mixin.build_form_success_response(form=form,
                                                          instance=user,
                                                          is_created=False)

        self.assertEqual(
            response,
            (200, {
                'user': user,
            }))

    def test_build_form_error_response_with_form(self):
        """Testing UpdateFormMixin.build_form_error_response with form"""
        form = MyForm(data={})
        response = self.mixin.build_form_error_response(form=form)

        self.assertEqual(
            response,
            (INVALID_FORM_DATA, {
                'fields': {
                    'username': ['This field is required.'],
                    'password': ['This field is required.'],
                },
            }))

    def test_build_form_error_response_with_validation_error(self):
        """Testing UpdateFormMixin.build_form_error_response with
        ValidationError
        """
        errors = ValidationError({
            'field1': ['This field is required.'],
            'field2': ['This is not a valid choice.'],
        })

        response = self.mixin.build_form_error_response(errors=errors)

        self.assertEqual(
            response,
            (INVALID_FORM_DATA, {
                'fields': {
                    'field1': ['This field is required.'],
                    'field2': ['This is not a valid choice.'],
                },
            }))

    def test_create_form(self):
        """Testing UpdateFormMixin.create_form"""
        form = self.mixin.create_form(
            data={
                'username': 'test-user',
                'password': 'test-pass',
            },
            request=None,
            form_kwargs={
                'custom': True,
            })

        self.assertTrue(form.custom)
        self.assertEqual(form.data['username'], 'test-user')
        self.assertEqual(form.data['password'], 'test-pass')
        self.assertEqual(form.data['first_name'], 'initial-first-name')

    def test_create_form_with_instance(self):
        """Testing UpdateFormMixin.create_form with instance"""
        user = User(username='test-user',
                    password='test-pass',
                    first_name='test-first-name',
                    last_name='test-last-name')

        form = self.mixin.create_form(
            data={
                'username': 'new-test-user',
                'last_name': 'new-last-name',
            },
            request=None,
            instance=user,
            form_kwargs={
                'custom': True,
            })

        self.assertTrue(form.custom)
        self.assertEqual(form.data['username'], 'new-test-user')
        self.assertEqual(form.data['password'], 'test-pass')
        self.assertEqual(form.data['first_name'], 'test-first-name')
        self.assertEqual(form.data['last_name'], 'new-last-name')

    def test_create_form_with_parse_field(self):
        """Testing UpdateFormMixin.create_form with parse_*_field method"""
        self.mixin.parse_username_field = \
            lambda value, my_state, **kwargs: 'test-%s-%s' % (value, my_state)

        form = self.mixin.create_form(
            data={
                'username': 'user',
                'password': 'test-pass',
            },
            request=None,
            form_kwargs={
                'custom': True,
            },
            my_state=123)

        self.assertTrue(form.custom)
        self.assertEqual(form.data['username'], 'test-user-123')
        self.assertEqual(form.data['password'], 'test-pass')

    def test_create_form_with_parse_field_raising_error(self):
        """Testing UpdateFormMixin.create_form with parse_*_field method
        raising ValidationError
        """
        def _parse_field(value, **kwargs):
            raise ValidationError('This field is just bad.')

        self.mixin.parse_username_field = _parse_field
        self.mixin.parse_password_field = _parse_field

        with self.assertRaises(ValidationError) as ctx:
            self.mixin.create_form(
                data={
                    'username': 'user',
                    'password': 'pass',
                },
                request=None,
                form_kwargs={
                    'custom': True,
                })

        self.assertEqual(ctx.exception.message_dict, {
            'username': ['This field is just bad.'],
            'password': ['This field is just bad.'],
        })

    def test_handle_form_request(self):
        """Testing UpdateFormMixin.handle_form_request"""
        self.spy_on(MyForm.save)

        response = self.mixin.handle_form_request(
            data={
                'username': 'test-user',
                'password': 'test-pass',
            },
            request=None,
            form_kwargs={
                'custom': True,
            },
            save_kwargs={
                'commit': False,
            },
            my_state=123)

        user = MyForm.save.last_call.return_value
        self.assertIsNone(user.pk)
        self.assertEqual(user.username, 'test-user')

        self.assertEqual(
            response,
            (201, {
                'user': user,
            }))

    def test_handle_form_request_with_instance(self):
        """Testing UpdateFormMixin.handle_form_request with instance"""
        user = User(username='test-user',
                    password='test-pass')

        response = self.mixin.handle_form_request(
            data={
                'username': 'new-user',
                'first_name': 'new-first-name',
                'last_name': 'new-last-name',
            },
            request=None,
            instance=user,
            form_kwargs={
                'custom': True,
            },
            my_state=123)

        user = User.objects.get()
        self.assertEqual(user.username, 'new-user')
        self.assertEqual(user.first_name, 'new-first-name')
        self.assertEqual(user.last_name, 'new-last-name')

        self.assertEqual(
            response,
            (200, {
                'user': user,
            }))

    def test_handle_form_request_with_error(self):
        """Testing UpdateFormMixin.handle_form_request with error"""
        response = self.mixin.handle_form_request(
            data={},
            request=None,
            form_kwargs={
                'custom': True,
            },
            my_state=123)

        self.assertEqual(
            response,
            (INVALID_FORM_DATA, {
                'fields': {
                    'username': ['This field is required.'],
                    'password': ['This field is required.'],
                },
            }))

    def test_handle_form_request_with_parse_field(self):
        """Testing UpdateFormMixin.handle_form_request with parse_*_field
        method
        """
        self.mixin.parse_username_field = \
            lambda value, my_state, **kwargs: 'test-%s-%s' % (value, my_state)

        self.spy_on(MyForm.save)

        response = self.mixin.handle_form_request(
            data={
                'username': 'user',
                'password': 'test-pass',
            },
            request=None,
            form_kwargs={
                'custom': True,
            },
            my_state=123)

        user = MyForm.save.last_call.return_value

        self.assertEqual(user.username, 'test-user-123')

        self.assertEqual(
            response,
            (201, {
                'user': user,
            }))

    def test_handle_form_request_with_parse_field_raising_error(self):
        """Testing UpdateFormMixin.handle_form_request with parse_*_field
        methods raising ValidationError
        """
        def _parse_field(value, **kwargs):
            raise ValidationError('This field is just bad.')

        self.mixin.parse_username_field = _parse_field
        self.mixin.parse_password_field = _parse_field

        response = self.mixin.handle_form_request(
            data={
                'username': 'user',
                'password': 'test-pass',
            },
            request=None,
            form_kwargs={
                'custom': True,
            },
            my_state=123)

        self.assertEqual(
            response,
            (INVALID_FORM_DATA, {
                'fields': {
                    'username': ['This field is just bad.'],
                    'password': ['This field is just bad.'],
                },
            }))

    def test_save_form(self):
        """Testing UpdateFormMixin.save_form"""
        form = MyForm(data={
            'username': 'test-user',
            'password': 'test-pass',
        })
        self.assertTrue(form.is_valid())

        user = self.mixin.save_form(form)

        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.pk, 1)
        self.assertEqual(User.objects.get(pk=1), user)

    def test_save_form_with_save_kwargs(self):
        """Testing UpdateFormMixin.save_form with save_kwargs"""
        form = MyForm(data={
            'username': 'test-user',
            'password': 'test-pass',
        })
        self.assertTrue(form.is_valid())

        user = self.mixin.save_form(form, save_kwargs={'commit': False})

        self.assertEqual(user.username, 'test-user')
        self.assertIsNone(user.pk)
        self.assertFalse(User.objects.exists())



"""Unit tests for djblets.forms.widgets.ListEditWidget."""

from __future__ import unicode_literals

from django import forms

from djblets.forms.widgets import ListEditWidget
from djblets.testing.testcases import TestCase


class ListEditWidgetTests(TestCase):
    """Unit tests for djblets.forms.widgets.ListEditWidget."""

    def test_render(self):
        """Testing ListEditWidget.render"""
        field = forms.CharField(widget=ListEditWidget())
        rendered = field.widget.render(
            name='my_field',
            value=' foo,  bar , baz ',
            attrs={
                'id': 'id_my_field',
                'class': 'my-value-class',
            })

        self.assertIn(
            '<div class="djblets-c-list-edit-widget list-edit-widget"'
            ' id="id_my_field_container">',
            rendered)
        self.assertIn('<input value="foo" type="text"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input value="bar" type="text"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input value="baz" type="text"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)

    def test_render_with_custom_separator(self):
        """Testing ListEditWidget.render with custom separator"""
        field = forms.CharField(widget=ListEditWidget(sep=';'))
        rendered = field.widget.render(
            name='my_field',
            value=' foo;  bar ; baz ',
            attrs={
                'id': 'id_my_field',
                'class': 'my-value-class',
            })

        self.assertIn(
            '<div class="djblets-c-list-edit-widget list-edit-widget"'
            ' id="id_my_field_container">',
            rendered)
        self.assertIn('<input value="foo" type="text"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input value="bar" type="text"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input value="baz" type="text"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)



"""Unit tests for API rate limiting."""

from __future__ import unicode_literals

import json

from django.test.utils import override_settings
from django.test.client import RequestFactory
from django.utils.encoding import force_text

from djblets.testing.testcases import TestCase
from djblets.webapi.errors import RATE_LIMIT_EXCEEDED
from djblets.webapi.resources.registry import unregister_resource
from djblets.webapi.resources.user import UserResource


class WebAPIRateLimitTests(TestCase):
    """Unit tests for API rate limiting."""

    def setUp(self):
        super(WebAPIRateLimitTests, self).setUp()

        self.factory = RequestFactory()
        self.user_resource = UserResource()

    def tearDown(self):
        super(WebAPIRateLimitTests, self).tearDown()

        unregister_resource(self.user_resource)

    @override_settings(API_ANONYMOUS_LIMIT_RATE='2/h')
    def test_api_rate_limit(self):
        """Testing API rate limiting."""
        request = self.factory.get('/api/')

        # First two should succeed
        response = self.user_resource(request)
        self.assertEqual(response.status_code, 200)

        response = self.user_resource(request)
        self.assertEqual(response.status_code, 200)

        # Third one should fail
        response = self.user_resource(request)
        self.assertEqual(response.status_code, 429)
        self.assertEqual(response['X-RateLimit-Limit'], '2')

        rsp = json.loads(force_text(response.content))
        self.assertEqual(rsp['err']['code'], RATE_LIMIT_EXCEEDED.code)
